package demux.ts;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import muxer.MuxManager;
import muxer.filter.TsStreamIdentifierMuxFilter;
import muxer.ts.M2tsRawMuxer;

import common.FileStream;
import common.Job;
import common.JobStreamI;
import common.StreamDescriptor;
import common.packet.Packet;
import common.ts.ProgramInfo;
import common.ts.TableHandler;
import common.ts.TsConstants;
import common.ts.TsTable;

import demux.Demuxer;

public /*abstract*/ class TsRawDemuxer extends Demuxer implements TsConstants {
	protected static final int READ_SIZE=100000;
	protected static final int PAT_PACKET=0;
	protected static final int PMT_PACKET=1;
	protected static final int PCR_PACKET=2;
	protected static final int STREAM_PACKET=3;
	
	private final int types[][]= {
			new int[] {0x1000, 0x10FF, StreamDescriptor.CLASSIFICATION_VIDEO},
			new int[] {0x1100, 0x11FF, StreamDescriptor.CLASSIFICATION_AUDIO},
			new int[] {0x1200, 0x12FF, StreamDescriptor.CLASSIFICATION_SUBTITLE}
	};
	
	protected int index;
	//protected ProgramInfo programInfo[];
	protected short selectedProgramNumbers[];
	protected int selectedProgramMapPids[];
	protected int emptyStreamDescriptorMaps;
	protected TableHandler tableHandler;
	
	protected int packetCount;
	protected int badPacketCount;
	
	public TsRawDemuxer() {
		tableHandler=new TableHandler();
		tableHandler.setMode(TableHandler.MODE_PARSE);
	}

	public void parse(JobStreamI stream, MuxManager muxManager) throws IOException {
		packetCount=badPacketCount=0;
		final int packetSize=getPacketSize();
		//stream.setChunkSize(packetSize);
		//64*1024 is the ECC size for Blu-Ray using this value ensures optimal Blu-Ray reads
		stream.setBufferChunks(packetSize, 64*1024);
		stream.open();
		try {
			int rc=MuxManager.CONTINUE;
			while (!stream.isEOS() && rc!=MuxManager.ABORT) {
				TsRawPacketPart packetPart=getPacketPart(stream);
				packetCount++;
				ByteBuffer buffer=packetPart.getByteBuffer();
				//packetPart.appendFrom(stream, packetSize);
				int offset=packetPart.getTsSyncOffset();
				if (buffer.get(offset)!=TS_SYNC_BYTE) {
					long position=-1;
					if (stream instanceof FileStream) {
						FileStream fileStream=(FileStream)stream;
						position=fileStream.getAbsoluteFilePosition() - buffer.capacity();
					}
					throw new IOException("Invalid or corrupt TS stream: " + position);
				}
				TsInfo info=packetPart.getTsInfo();
				
				
				if (info.error) {
					badPacketCount++;
				} else {
					TsRawPacketPart parts[]=tableHandler.append(packetPart);
					if (parts !=null) {
						for (int i=0;i<parts.length && rc !=MuxManager.ABORT;i++) {
							//Packet type is not 100% accurate
							rc=addPacketPart(parts[i], PAT_PACKET, muxManager);
						}
					} else if (tableHandler.isPcrPid(info.pid)) {
						
						packetPart.setSystemClockReference(info.adaptationControl >= 0x20);
						int packetType=PCR_PACKET;
						if ((info.adaptationControl & 0x10)==0x10 & isCopied(info.pid)) {
							packetPart.setAvStream(true);
							packetType=STREAM_PACKET;
						}
						rc=addPacketPart(packetPart, packetType, muxManager);
					} else if (isCopied(info.pid)) {
						packetPart.setAvStream(true);
						rc=addPacketPart(packetPart, STREAM_PACKET, muxManager);
					}
				}
			}
			
		} finally {
			stream.close();
		}
		streamEnded(muxManager);
	}

	private int addPacketPart(final TsRawPacketPart packetPart, final int packetType, final MuxManager muxManager) throws IOException {
		Packet packet=new Packet(packetPart.getTsInfo().pid,index++);
		packet.setPayload(packetPart);
		return addPacket(packet, packetType, muxManager);
	}
	protected void streamEnded(MuxManager muxManager) throws IOException {
		muxManager.streamEnded();
	}
	
	protected int getPacketSize() {
		return TS_PACKET_SIZE;
	}
	public Map<Integer, StreamDescriptor> getStreamDescriptorMap(JobStreamI stream, MuxManager muxManager) throws IOException {
		TsStreamIdentifierMuxFilter identifier=new TsStreamIdentifierMuxFilter(getPacketSize()*READ_SIZE, this);
		//PacketBuffer packetBuffer=new PacketBuffer(regulator);
		muxManager.add(identifier);
		muxManager.mux(stream);
		stream.close();
		stream.reset();
		muxManager.remove(identifier);
		Map<Integer, StreamDescriptor> map=identifier.getStreamDescriptorMap();
		//Use BluRay defaults if unknown
		for (Iterator sdi=map.values().iterator();sdi.hasNext();) {
			StreamDescriptor streamDescriptor=(StreamDescriptor)sdi.next();
			if (streamDescriptor.getClassification()==StreamDescriptor.CLASSIFICATION_UNKNOWN) {
				int pid=streamDescriptor.getStreamId();
				for (int j=0;j<types.length;j++) {
					if (pid >= types[j][0] && pid <= types[j][1]) {
						streamDescriptor.setClassification(types[j][2]);
					}
				}
				
			}
		}
		return map;
	}	
	/**
	 * 
	 * @param programNumbers
	 */

	public void setStreams(List<StreamDescriptor> streamDescriptorList) {
		super.setStreams(streamDescriptorList);
		HashSet<Short> set=new HashSet<Short>();
		for (int i=0;i<streamDescriptorList.size();i++) {
			set.add(Short.valueOf(streamDescriptorList.get(i).getProgramNumber()));
		}
		short programNumbers[]=new short[set.size()];
		int j=0;
		for (Iterator<Short> i=set.iterator();i.hasNext();) {
			Short s=i.next();
			programNumbers[j++]=s.shortValue();
		}
		tableHandler.setSelectedProgramNumbers(programNumbers);
		tableHandler.setStreamList(streamDescriptorList);
	}
	
	public void setSelectedProgramNumber(short programNumber) {
		ProgramInfo programInfo[]=tableHandler.getProgramInfo();
		for (int i=0;i<programInfo.length;i++) {
			if (programInfo[i].getProgramNumber()==programNumber) {
				super.setStreams(new ArrayList(programInfo[i].getStreamDescriptorMap().values()));
			}
		}
		tableHandler.setSelectedProgramNumbers(new short[]{programNumber});
	}
	public static void main(String args[]) {
		/*String titleName="F:/T2.ts";
		
		try {
			TsRawDemuxer demuxer=new TsRawDemuxer(new File(titleName));
			MuxManager muxManager=new MuxManager(demuxer);
			Map map=muxManager.getStreamDescriptorMap();
			System.out.println(map);
			muxManager.add(new M2tsRawMuxer(new File("F:/Black Hawk Down.m2ts")));
			muxManager.mux();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}*/
		String titleName="F:/HD DVD/test_avc.ts";
		
		try {
			TsRawDemuxer demuxer=new TsRawDemuxer();
			FileStream fileJob=new FileStream(new Job(null), new File[]{new File(titleName)});
			MuxManager muxManager=new MuxManager(demuxer);
			Map map=muxManager.getStreamDescriptorMap(fileJob);
			System.out.println(map);
			muxManager.add(new M2tsRawMuxer(new FileOutputStream("F:/test_avc.m2ts").getChannel()));
			muxManager.mux(fileJob);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	//protected abstract int addPacket(Packet packet, int packetType, MuxManager muxManager) throws IOException;
	protected int addPacket(Packet packet, int packetType, MuxManager muxManager) throws IOException {
		return muxManager.addPacket(packet);
	}

	
	protected TsRawPacketPart getPacketPart(JobStreamI stream) throws IOException{
		return new TsRawPacketPart(stream.slice(TS_PACKET_SIZE));
	}
	public ProgramInfo[] getProgramInfo() {
		return tableHandler.getProgramInfo();
	}

	public TableHandler getTableHandler() {
		return tableHandler;
	}

	public int getPacketCount() {
		return packetCount;
	}

	public int getBadPacketCount() {
		return badPacketCount;
	}
}
