package muxer.ts;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.WritableByteChannel;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;

import muxer.MuxFilter;
import muxer.MuxManager;
import muxer.MuxUtil;
import muxer.PacketBuffer;
import regulator.StreamPtsPacketCountRegulator;

import common.FileStream;
import common.Job;
import common.MuxBuffer;
import common.StreamDescriptor;
import common.Util;
import common.mpeg.MpegConstants;
import common.mpeg.MpegUtil;
import common.packet.ClockReferenceI;
import common.packet.Packet;
import common.packet.PacketPart;
import common.packet.PesPacket;
import common.packet.PesPacketPart;

import demux.ts.TsRawPacketPart;
import demux.ty.TyPesDemuxer;

public class TsPesMuxer extends MuxFilter implements MpegConstants{
	private final long SCR_INTERVAL= PTS_FREQ/10 - PTS_FREQ/100;//90ms
	//private OutputStream out;
	private int syncStreamId;
	private long lastPcr=Long.MAX_VALUE;
	private final long pcrOffset=PTS_FREQ/100;//10 millis
	//private final long pcrOffset=PTS_FREQ/4;//250 millis
	private final Map<Integer, Integer> streamPidMap=new TreeMap<Integer, Integer>();
	private final TsRawPacketPart patBuffer;
	private final TsRawPacketPart pmtBuffer;
	private final TsRawPacketPart pcrBuffer;
	private final byte streamIdBuffer[]={90, (byte)0xF2, 0x15, (byte)0xF0, 0x00};
	private final HashMap<Integer, int[]> continuityCountMap=new HashMap<Integer, int[]>();
	
	private TsPacketWriter packetWriter;
	
	public TsPesMuxer(WritableByteChannel out) {
		this.packetWriter=getPacketWriter(out);
		patBuffer=new TsRawPacketPart(TS_PAT);
		
		pmtBuffer=new TsRawPacketPart(TS_PMT_HEADER);
		
		pcrBuffer=new TsRawPacketPart(TS_PCR);
	}
	
	protected void init() throws IOException {
		Collection<StreamDescriptor> muxedStreamDescriptorColl=muxManager.getMuxedStreamDescriptorColl();
		syncStreamId=MuxUtil.getLinearStream(muxedStreamDescriptorColl).getStreamId();
		PacketBuffer packetBuffer=muxManager.getPacketBuffer();
		StreamPtsPacketCountRegulator regulator=new StreamPtsPacketCountRegulator(new int[]{syncStreamId}, 6);
		if (packetBuffer==null) {
			muxManager.add(packetBuffer=new PacketBuffer(regulator));
		} else {
			packetBuffer.addRegulator(regulator);
		}
		TreeMap<Integer, StreamDescriptor> pidStreamMap=new TreeMap<Integer, StreamDescriptor>();
		
		//Use BluRay mappings
		int audioStream=0x1100;
		int subtitleStream=0x1200;
		int otherStream=0x1400;
		for (StreamDescriptor streamDescriptor : muxedStreamDescriptorColl) {
			Integer pid;
			switch (streamDescriptor.getClassification()) {
				case StreamDescriptor.CLASSIFICATION_VIDEO:
					pid=0x1011;
				case StreamDescriptor.CLASSIFICATION_AUDIO:
					pid=audioStream++;
					break;
				case StreamDescriptor.CLASSIFICATION_SUBTITLE:
					pid=subtitleStream++;
					break;
				default:
					pid=otherStream++;
			}
			streamPidMap.put(streamDescriptor.getStreamId(), pid);
			pidStreamMap.put(pid, streamDescriptor);
			continuityCountMap.put(pid, new int[1]);
		}
		packetWriter.setSyncStreamId(((Integer)streamPidMap.get(syncStreamId)).intValue());
		
		for (Integer streamId : pidStreamMap.keySet()) {
			StreamDescriptor streamDescriptor=(StreamDescriptor)pidStreamMap.get(streamId);
			streamIdBuffer[0]=streamDescriptor.getTsType();
			int value=streamId.intValue() | 0xE000;
			Util.set16(streamIdBuffer,1,value);
			streamIdBuffer[3]=(byte)0xF0;
			byte[] esDescriptorBuffer=streamDescriptor.getTsEsDescriptor();
			streamIdBuffer[4]=(byte)esDescriptorBuffer.length;
			pmtBuffer.appendFrom(streamIdBuffer, 0, streamIdBuffer.length);
			pmtBuffer.appendFrom(esDescriptorBuffer, 0, esDescriptorBuffer.length);
		}
		int sectionLen=(pmtBuffer.limit()-8/*Ts Header*/+4/*Checksum*/)| 0xB000;
		pmtBuffer.set16(sectionLen, 6);
		long crc=MpegUtil.calculateCRC32(pmtBuffer, 5, pmtBuffer.limit()-5);
		byte crcBuffer[]=new byte[4];
		Util.set32(crcBuffer, 0, crc);
		pmtBuffer.appendFrom(crcBuffer, 0, crcBuffer.length);
		padPacket(pmtBuffer);
		
		padPacket(patBuffer);
		packetWriter.writeTsPacket(patBuffer);
		packetWriter.writeTsPacket(pmtBuffer);
		
		padPacket(pcrBuffer);//Don't write yet
		pcrBuffer.getTsInfo();//Prime TS info
		
	}	
	
	protected void padPacket(MuxBuffer buffer) {
		while (buffer.limit() < 188) {
			buffer.put((byte)0xFF);
		}
	}
	public int addPacket(Packet packet) throws IOException {
		long pts=packet.getClockReference(ClockReferenceI.CLOCK_REFERENCE_TYPE_PTS);
		try {
			if (packet.getStreamId()==syncStreamId && pts != ClockReferenceI.CLOCK_REFERENCE_NONE) {
				//long pts=packet.getClockReference(ClockReferenceI.CLOCK_REFERENCE_TYPE_PTS);
				if (lastPcr==Long.MAX_VALUE) {
					lastPcr=pts - pcrOffset;
					pcrBuffer.setSystemClockReference(lastPcr);
					packetWriter.writeTsPacket(pcrBuffer);
				} else if (pts - lastPcr > SCR_INTERVAL) {
					long nextPcr=pts - pcrOffset  ;
					pcrBuffer.setSystemClockReference(nextPcr);
					lastPcr=nextPcr;
					packetWriter.writeTsPacket(patBuffer);
					packetWriter.writeTsPacket(pmtBuffer);
					packetWriter.writeTsPacket(pcrBuffer);
				}
				//System.out.println(Integer.toHexString(packet.getStreamId()) + ": " + MpegUtil.pesTimeToString(lastPcr) + " - " + MpegUtil.pesTimeToString(pts) + " - " + MpegUtil.pesTimeToString(pts-lastPcr));
			}
			Integer pid=streamPidMap.get(packet.getStreamId());
			int continuityCount[]=continuityCountMap.get(pid);
			PacketPart payload=packet.getPayload();
			TsRawPacketPart tmpBuffer=packetWriter.getPacketPart();
			if (packet instanceof PesPacket) {
				//This should always be true
				ByteBuffer byteBuffer=tmpBuffer.getByteBuffer();
				byteBuffer.limit(4);
				byteBuffer.put(0,TS_SYNC);
				PesPacket pesPacket=(PesPacket)packet;
				PesPacketPart header=pesPacket.getHeader();
				int tsPid=pid.intValue() | 0x4000;//Payload start
				byteBuffer.putShort(1, (short)tsPid);
				tmpBuffer.setAvStream(true);
				int len=header.get16(4);
				if (len==0) {
					len=header.limit() + payload.limit() - 6;
					if (len > 0xffff) {
						//System.out.println("Doah! packet too big: " + len);
						len=0;
					}
					header.set16(len, 4);
				}
				//byteBuffer.put((byte)(0x10 + continuityCount[0]++));
				byteBuffer.position(4);
				tmpBuffer.appendFrom(header.getByteBuffer(), 0, header.limit());
				
			} else {
				System.out.println("Doah! This muxer only supports PesPackets");
			}
			int bytesRemaining=payload.limit();
			final int packetSize=188;
			for (int offset=0;bytesRemaining > 0;) {
				int packetRemaining=packetSize-tmpBuffer.limit();
				int adapationControl;
				if (bytesRemaining > packetRemaining) {
					adapationControl=0x10;
					tmpBuffer.appendFrom(payload.getByteBuffer(), offset, packetRemaining);
				} else {
					adapationControl=0x30;
					packetRemaining-=2;//Subtract off 2 for the adaptation field
					int bytesToCopy=Math.min(bytesRemaining, packetRemaining);
					int offsetSize=packetRemaining-bytesToCopy;
					tmpBuffer.put((byte)(offsetSize+1));
					tmpBuffer.put(B0);
					for (int i=0;i<offsetSize;i++) {
						tmpBuffer.put((byte)0xFF);
					}
					tmpBuffer.appendFrom(payload.getByteBuffer(), offset, bytesToCopy);
					/*if (tmpBuffer.getLen()!=188) {
						throw new RuntimeException("Adaptation Logic failed");
					}*/
				}
				bytesRemaining-=packetRemaining;
				offset+=packetRemaining;
				continuityCount[0]=(continuityCount[0]+1) & 0xf;
				tmpBuffer.setByte((byte)(adapationControl + continuityCount[0]), 3);
				packetWriter.writeTsPacket(tmpBuffer);
				
				tmpBuffer=packetWriter.getPacketPart();
				tmpBuffer.setAvStream(true);
				ByteBuffer byteBuffer=tmpBuffer.getByteBuffer();
				byteBuffer.limit(4);
				byteBuffer.put(0,TS_SYNC);
				byteBuffer.putShort(1,(short)pid.intValue());
				byteBuffer.position(4);
			}
		} catch (IOException ioe) {
			packetWriter.close();
			throw ioe;
		}
		return super.addPacket(packet);
	}
	public void streamEnd() throws IOException {
		packetWriter.streamEnd();
		super.streamEnd();
	}
	protected TsPacketWriter getPacketWriter(WritableByteChannel out) {
		return new TsPacketWriter(out);
	}
	public static void main(String args[]) {
		try {
			//String fileName="The Magic School Bus - The Magic School Bus Goes on Air";
			//String fileName="CSI_ Crime Scene Investigation - Monster in the Box";
			String fileName="Tombstone";
			System.out.println("Starting");
			String titleName="F:\\test Tys\\"+fileName+".ty";
			FileStream fileJob=new FileStream(new Job(null), new File[]{new File(titleName)});
			TyPesDemuxer demuxer=new TyPesDemuxer();
			//new RestreamingPesMuxer(buffer, new NullOutputStream());
			long start=System.currentTimeMillis();
			//in.skip(10631512064l);
			MuxManager muxManager=new MuxManager(demuxer);
			Map map=muxManager.getStreamDescriptorMap(fileJob);
			System.out.println(map);
			PacketBuffer packetBuffer=muxManager.getPacketBuffer();
			//packetBuffer.setRegulator(PacketBuffer.ALWAYS_FULL);
			//muxManager.remove(packetBuffer);
			muxManager.add(new TsPesMuxer(new FileOutputStream("C:/"+fileName+".ts").getChannel()));
			muxManager.mux(fileJob);
			System.out.println("Done: " + (System.currentTimeMillis()-start));
		} catch (IOException e) {
			System.out.println(e);
		}
	}
	public int getPosition() {
		return POSITION_MUXER;
	}
}
