package util;

import java.io.File;
import java.io.FileFilter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import muxer.MuxManager;
import ui.tonmt.ToNMTPrefs;

import common.AudioStreamDescriptor;
import common.ExternalProcess;
import common.FileStream;
import common.InputStreamLogger;
import common.Job;
import common.StreamDescriptor;
import common.StreamIdentifier.DolbyDigitalAnalzyer;
import common.StreamIdentifier.DolbyTrueHdAnalyzer;
import common.StreamIdentifier.DtsAnalyzer;
import common.StreamIdentifier.DtsMasterAudioAnalyzer;
import common.StreamIdentifier.LpcmAnalyzer;

//import deadbeef.SupTools.Core;
//import deadbeef.SupTools.CoreException;
import demux.Demuxer;
import demux.Eac3toDemuxer;

public class Eac3toTsMuxerConverter extends StreamConverter {
	public static final String EAC3TO_POSTFIX="eac3toPostfix";
	public static final String KEEP_TEMP_FILES="keepTempFiles";
	
	public static final String HD_DVD_EXTENSION="1.EVO";
	public static final String NEXT_EVO_EXTENSION="2.EVO";
		
	private static final String EAC3TO_EXT_LPCM=".wav";
	private static final String EAC3TO_EXT_DTS=".dts";
	private static final String EAC3TO_EXT_TRUEHD=".thd";
	private static final String EAC3TO_EXT_AC3=".ac3";
	private static final String EAC3TO_EXT_TRUEHD_AC3=".thd+ac3";
	private static final String EAC3TO_KEEP_DIALOG_NORMALIZATION="-keepDialnorm";
	
	public static final String TS_MUXER_AC3="A_AC3";
	public static final String TS_MUXER_DTS="A_DTS";
	public static final String TS_MUXER_LPCM="A_LPCM";
	public static final String TS_MUXER_EAC3=TS_MUXER_AC3;
	public static final String TS_MUXER_TRUEHD_AC3=TS_MUXER_AC3;
	
	
	public static final HashMap<String, String> extToTsMuxerMap=new HashMap<String, String>();
	
	static {
		extToTsMuxerMap.put(EAC3TO_EXT_AC3, TS_MUXER_AC3);
		extToTsMuxerMap.put(EAC3TO_EXT_DTS, TS_MUXER_DTS);
		extToTsMuxerMap.put(".dma", TS_MUXER_DTS);
		extToTsMuxerMap.put(EAC3TO_EXT_LPCM, TS_MUXER_LPCM);
		extToTsMuxerMap.put(".eac3", TS_MUXER_EAC3); //Technically not legal
		extToTsMuxerMap.put(EAC3TO_EXT_TRUEHD_AC3, TS_MUXER_TRUEHD_AC3);
		extToTsMuxerMap.put(".h264", "V_MPEG4/ISO/AVC");
		extToTsMuxerMap.put(".mpeg2", "V_MPEG-2");
		extToTsMuxerMap.put(".vc1", "V_MS/VFW/WVC1");
		extToTsMuxerMap.put(".sup", "S_HDMV/PGS");

	}
	
	private ExternalProcess process; 

	public Eac3toTsMuxerConverter(ToNMTPrefs prefs) {
		Demuxer demuxer=new Eac3toDemuxer(prefs.getEac3toPath());
		muxManager=new MuxManager(demuxer);
		this.prefs=prefs;
	}
	public File[] findFiles(String fileName) {
		File files[];
		fileName=fileName.toUpperCase();
		if (fileName.endsWith(HD_DVD_EXTENSION)) {
			File nextFile=new File(fileName.substring(0,fileName.length()-HD_DVD_EXTENSION.length()) + NEXT_EVO_EXTENSION);
			if (nextFile.exists()) {
				logger.info("Detected and added: " + nextFile.getAbsolutePath());
				files=new File[] {new File(fileName),nextFile};
			} else {
				files=new File[] {new File(fileName)};
			}
		} else {
			files=new File[] {new File(fileName)};
		}
		return files;
	}
	public void convert(FileStream fileStream, String outputFileName, List<StreamDescriptor> list) throws IOException {
		ArrayList<String> fileNameList=new ArrayList<String>(20);
		ArrayList<String> cmdList=new ArrayList<String>(20);
		
		cmdList.add(prefs.getEac3toPath());
		File files[]=fileStream.getFiles();
		StringBuffer sb=new StringBuffer();
		String sep="";
		for (int i=0;i<files.length;i++) {
			sb.append(sep);
			sep="+";
			sb.append('"');
			sb.append(files[i].getAbsolutePath());
			sb.append('"');
		}
		cmdList.add(sb.toString());
		
		StreamDescriptor video=getSubList(list, StreamDescriptor.CLASSIFICATION_VIDEO).get(0);
		String videoFileName='"'+prefs.getDemuxPrefix()+video.getFileNameExtension()+'"';
		cmdList.add(video.getStreamId()+":");
		cmdList.add(videoFileName);
		
		//int audioCmdStart=cmdList.size();
		fileNameList.add(videoFileName);
		
		AudioStreamDescriptor lpcm=null;
		AudioStreamDescriptor hd=null;
		boolean sdPresent=false;
		
		List<StreamDescriptor> audioList=getSubList(list, StreamDescriptor.CLASSIFICATION_AUDIO);
		for (StreamDescriptor audio : audioList) {
			AudioStreamDescriptor asd=(AudioStreamDescriptor)audio;
			String ext;
			String codec=asd.getCodec();
			boolean isTrueHd=codec.startsWith(DolbyTrueHdAnalyzer.CODEC_NAME);
			String originalExt=asd.getFileNameExtension();
			
			if (codec.startsWith(DolbyDigitalAnalzyer.EAC3_CODEC_NAME)) {
				ext=getExtensionByAction(originalExt,prefs.getEAC3Action());
				hd=asd;
			} else if (isTrueHd) {
				ext=getExtensionByAction(originalExt,prefs.getTrueHDAction());
				hd=asd;
			} else if (codec.startsWith(DtsAnalyzer.CODEC_NAME)) {
				ext=getExtensionByAction(originalExt,prefs.getDtsAction());
			} else if (codec.startsWith(DolbyDigitalAnalzyer.AC3_CODEC_NAME) || 
					codec.startsWith(LpcmAnalyzer.CODEC_NAME) ||
					codec.startsWith(DtsMasterAudioAnalyzer.CODEC_NAME)) {
				ext=originalExt;
			} else {
				logger.warn("Unknown audio codec: " + asd);
				continue;
			}
			if (lpcm==null && EAC3TO_EXT_LPCM.equals(ext)) {
				lpcm=asd;
			}
			sdPresent|=EAC3TO_EXT_AC3.equals(ext) || EAC3TO_EXT_DTS.equals(ext);
			String fileName='"'+prefs.getDemuxPrefix()+asd.getStreamId()+ext+'"';
			fileNameList.add(fileName);
			cmdList.add(asd.getStreamId()+":");
			if (EAC3TO_EXT_TRUEHD_AC3.equals(ext)) {
				cmdList.add(fileName);
				cmdList.add("-" + prefs.getTrueHdCoreRate());
			} else {
				cmdList.add(fileName);
			}
			//Removing Dialog Normalization causes problems on Denon Receivers
			//So if we aren't transcoding the audio, don't mess with it
			if (prefs.isKeepDialogNormalization() || (ext.equals(originalExt) || EAC3TO_EXT_TRUEHD_AC3.equals(ext))) {
				cmdList.add(EAC3TO_KEEP_DIALOG_NORMALIZATION);
			}
		}
		
		if (!sdPresent) {
			if (prefs.isEnsureSDAudio()) {
				String type=(lpcm != null) ? ('.' + prefs.getConvertToType().toLowerCase()) : EAC3TO_EXT_AC3;
				String fileName='"'+prefs.getDemuxPrefix() + hd.getStreamId() + type +'"';
				
				fileNameList.add(1,fileName);
				cmdList.add(/*audioCmdStart,*/hd.getStreamId()+":");
				cmdList.add(/*audioCmdStart+1,*/fileName);
				
			} else if (lpcm!=null) {
				//NMT Bug work-around: If this is LPCM and no SD audio codec is present, add an AC3
				String fileName='"'+prefs.getDemuxPrefix()+".ac3"+'"';
				fileNameList.add(fileName);
				
				cmdList.add(lpcm.getStreamId()+":");
				cmdList.add(fileName);
				cmdList.add("-192");
				cmdList.add("-down2");
			} 
		}
		
		List<StreamDescriptor> subtitleList=getSubList(list, StreamDescriptor.CLASSIFICATION_SUBTITLE);
		List<String> subtitleNameList=new ArrayList<String>(subtitleList.size());
		for (StreamDescriptor subtitle : subtitleList) {
			cmdList.add(subtitle.getStreamId()+":");
			String fileName=prefs.getDemuxPrefix()+subtitle.getStreamId() + ".sup";
			subtitleNameList.add(fileName);
			cmdList.add('"' + fileName + '"');
			//This will be the filename after the second pass
			fileNameList.add('"' + fileName + "_bd.sup" + '"');
		}
		
		cmdList.add("-stripPulldown");
		cmdList.add("-progressNumbers");
		
		String eac3toPostfix=System.getProperty(EAC3TO_POSTFIX);
		if (eac3toPostfix != null) {
			String eac3toPostfixArray[]=stripQuotes(eac3toPostfix).split(" ");
			for (int i=0;i<eac3toPostfixArray.length;i++) {
				cmdList.add(eac3toPostfixArray[i]);
			}
		}
		// "analyze: "
		// "process: "
		String cmdArray[]=new String[cmdList.size()];
		cmdList.toArray(cmdArray);
		process=new ExternalProcess(fileStream.getJob(),"process: ","analyze: ", new InputStreamLogger());
		int rc=process.start(cmdArray);
		if (rc!=0) {
			process=null;
			throw new IOException("Demux Streams Failed");
		}
		
		//Process subtitles
		/*for (String subtitleName : subtitleNameList) {
			logger.info("Converting subtitle: " + subtitleName);
	        try {
	            Core.setOutputMode(Core.OutputMode.BDSUP);
	            Core.setFPSTrg(Core.FPS_24P);            
	            Core.Resolution r=Core.Resolution.HD_1080;
	            Core.init(null);
	            Core.readSup(subtitleName);
	            Core.setOutputResolution(r);
	            Core.scanSubtitles();
	            Core.writeSub(subtitleName+"_bd.sup");
	            Core.exit();
	        } catch (CoreException e) {
	        	logger.error(e);
	        }			
		}*/
		
		//Write the meta file for TsMuxeR
		String metaFileName=prefs.getDemuxPrefix() + ".meta";
		PrintStream out=new PrintStream(new FileOutputStream(metaFileName));
		out.println("MUXOPT --no-pcr-on-video-pid --new-audio-pes --vbr");
		long length=0;
		for (int i=0;i<fileNameList.size();i++) {
			String fileName=fileNameList.get(i);
			String fileNameNoQuote=stripQuotes(fileName);
			File file=new File(fileNameNoQuote);
			String extension=getExtension(fileNameNoQuote);
			String tsMuxerLine;
			if (!file.exists()) {
				file=findDelayFile(file, extension);
				if (file == null) {
					throw new RuntimeException("Failed to find file: " + fileNameNoQuote);
				}
				fileNameNoQuote=file.getAbsolutePath();
				String delay=fileNameNoQuote.substring(fileNameNoQuote.lastIndexOf(' ')+1, fileNameNoQuote.length()-extension.length());
				fileName='"'+fileNameNoQuote + '"';
				fileNameList.set(i, fileName);
				tsMuxerLine=extToTsMuxerMap.get(extension) +","+ fileName +",timeshift="+delay;
			} else {
				tsMuxerLine=extToTsMuxerMap.get(extension) +","+ fileName;
			}
			length+=file.length();
			out.println(tsMuxerLine);
		}
		
		out.close();
		//"0.1% complete"
		Job job=fileStream.getJob();
		job.setSize((int)(length/Job.KB));
		job.setDescription("Muxing: " + outputFileName);
		process=new ExternalProcess(fileStream.getJob()," complete", null, new InputStreamLogger());
		rc=process.start(new String[]{new File(prefs.getTsMuxerPath()).getAbsolutePath(),metaFileName, outputFileName});
		process=null;
		
		if (System.getProperty(KEEP_TEMP_FILES)==null) {
			for (int i=0;i<fileNameList.size();i++) {
				new File(stripQuotes(fileNameList.get(i))).delete();
			}
			new File(metaFileName).delete();
			for (String subtitleName : subtitleNameList) {
				new File(subtitleName).delete();
			}
		}
	}
	private File findDelayFile(File file, final String extension) {
		String fileNameNoQuotes=file.getAbsolutePath();
		final String prefix=fileNameNoQuotes.substring(0, fileNameNoQuotes.length()-extension.length());
		
		FileFilter fileFilter=new FileFilter() {
			public boolean accept(File file) {
				String fileNameNoQuotes=file.getAbsolutePath();
				
				return fileNameNoQuotes.startsWith(prefix) && fileNameNoQuotes.endsWith(extension);
			}
		};
		File files[]=file.getParentFile().listFiles(fileFilter);
		if (files.length==1) {
			return files[0];
		} else {
			return null;
		}
	}
	private String getExtensionByAction(String orgExt, String action) {
		if (ToNMTPrefs.ACTION_PASSTHROUGH.equals(action)) {
			return orgExt;
		} else if (ToNMTPrefs.ACTION_ADD_AC3.equals(action)) {
			return EAC3TO_EXT_TRUEHD_AC3;
		} else if (ToNMTPrefs.ACTION_CONVERT.equals(action)) {
			String codec=prefs.getConvertToType();
			if (ToNMTPrefs.AUDIO_TYPE_LPCM.equals(codec)) {
				return EAC3TO_EXT_LPCM;
			} else if (ToNMTPrefs.AUDIO_TYPE_TRUEHD.equals(codec)) {
				return EAC3TO_EXT_TRUEHD;
			} else {
				return "." + codec.toLowerCase();
			}
		} else {
			throw new RuntimeException("Unknown Action: " + action);
		}
	}
	private static String stripQuotes(String fileName) {
		if (fileName.charAt(0)=='"') {
			return fileName.substring(1,fileName.length()-1);
		} else {
			return fileName;
		}
	}
	private String getExtension(String fileName) {
		fileName=stripQuotes(fileName);
		return fileName.substring(fileName.lastIndexOf('.'));
	}
	public void abort() {
		if (process !=null) {
			process.abort();
		}
	}
}
